'''`bestdori.characters`

BanG Dream! 角色相关操作'''
from typing import Optional, Literal, Any

from .post import get_list
from .utils import hex_to_rgb
from .utils.utils import API, RES, ASSETS
from .utils.network import Api, Res, Assets
from .exceptions import (
    CharacterNotExistError
)

# 获取总角色信息
def get_all(index: Literal[2]=2, proxy: Optional[str]=None) -> dict[str, dict[str, Any]]:
    '''获取总角色信息

    参数:
        index (Literal[2], optional): 指定获取哪种 `all.json`
            `2`: 获取所有已有角色信息 `all.2.json`
        
        proxy (Optional[str], optional): 代理服务器

    返回:
        dict[str, dict[str, Any]]: 获取到的总角色信息
    '''
    return Api(API['characters']['all'].format(index), proxy=proxy).request('get').json()

# 获取主要角色信息
def get_main(index: Literal[3]=3, proxy: Optional[str]=None) -> dict[str, dict[str, Any]]:
    '''获取主要角色信息

    参数:
        index (Literal[3], optional): 指定获取哪种 `all.json`
            `3`: 获取所有已有主要角色信息 `all.3.json`
        
        proxy (Optional[str], optional): 代理服务器

    返回:
        dict[str, dict[str, Any]]: 获取到的主要角色信息
    '''
    return Api(API['characters']['main'].format(index), proxy=proxy).request('get').json()

# 角色类
class Character:
    '''角色类

    参数:
        id_ (int): 角色 ID
        
        proxy (Optional[str], optional): 代理服务器
    '''
    # 初始化
    def __init__(self, id_: int, proxy: Optional[str]=None) -> None:
        '''角色类

        参数:
            id_ (int): 角色 ID
            
            proxy (Optional[str], optional): 代理服务器
        '''
        self.id: int = id_
        '''角色 ID'''
        self._info: dict[str, Any] = {}
        '''角色信息'''
        self.proxy: Optional[str] = proxy
        '''代理服务器'''
        # 检测 ID 是否存在
        all_id = get_all(2, proxy=proxy)
        if not str(id_) in all_id.keys():
            raise CharacterNotExistError(id_)
        return
    
    # 获取角色信息
    def get_info(self) -> dict[str, Any]:
        '''获取角色信息

        返回:
            dict[str, Any]: 角色详细信息
        '''
        if len(self._info) <= 0:
            # 如果没有角色信息存储
            response = Api(
                API['characters']['info'].format(self.id), proxy=self.proxy
            ).request('get')
            self._info = dict(response.json())
        return self._info
    
    # 获取角色评论
    def get_comment(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_ASC'
    ) -> dict[str, Any]:
        '''获取角色评论

        参数:
            limit (int, optional): 展示出的评论数，默认为 20
            
            offset (int, optional): 忽略前面的 `offset` 条评论，默认为 0
            
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 排序顺序，默认时间顺序

        返回:
            dict[str, Any]: 搜索结果
            ```python
            result: bool # 是否有响应
            count: int # 搜索到的评论总数
            posts: list[dict[str, Any]] # 列举出的评论
            ```
        '''
        return get_list(
            proxy=self.proxy,
            category_name='CHARACTER_COMMENT',
            category_id=str(self.id),
            order=order,
            limit=limit,
            offset=offset
        )
    
    # 获取角色名称
    @property
    def name(self) -> str:
        '''获取角色名称

        返回:
            str: 角色名称
        '''
        info = self.get_info()
        # 获取 characterName 数据
        if (character_name := info.get('characterName', None)) is None:
            raise Exception('无法获取角色名称。')
        # 获取第一个非 None 角色名称
        try:
            return next(filter(lambda x: x is not None, character_name))
        except StopIteration:
            raise Exception('无法获取角色名称。')
    
    # 获取角色图标
    @property
    def icon(self) -> bytes:
        '''获取角色图标

        返回:
            bytes: 角色图标字节数据 `bytes`
        '''
        return Res(RES['icon']['png'].format(name=f'chara_icon_{self.id}')).get()
    
    # 获取角色颜色
    @property
    def color(self) -> tuple[int, int, int]:
        '''获取角色颜色

        返回:
            tuple[int, int, int]: 角色颜色元组
        '''
        info = self.get_info()
        # 获取 colorCode 数据
        if (color_code := info.get('colorCode', None)) is None:
            raise Exception('无法获取角色颜色。')
        # 将 colorCode 转换为颜色元组
        try:
            return hex_to_rgb(color_code)
        except ValueError:
            raise Exception('无法获取角色颜色。')
    
    # 获取角色主视觉图
    def get_kv_image(self) -> bytes:
        '''获取角色主视觉图

        返回:
            bytes: 主视觉图资源字节 `bytes`
        '''
        return Assets(
            ASSETS['characters']['character_kv_image'].format(id=self.id), 'jp', self.proxy
        ).get()



'''`bestdori.ayachan.utils.network`

向 ayachan 发送请求相关模块'''
from json import dumps
from io import BufferedReader
from httpx import Response, Request, Client
from typing import Optional, Literal, Union, cast, Any

# 向 ayachan 发送 API 请求类
class Api:
    '''向 ayachan 发送 API 请求类

    参数:
        api (str): 请求的 API 地址
        
        proxy (Optional[str]): 代理服务器'''
    api: str
    '''请求的 API 地址'''
    proxy: Optional[str]=None
    '''代理服务器'''
    # 初始化
    def __init__(
        self,
        api: str,
        proxy: Optional[str]=None
    ) -> None:
        '''初始化'''
        self.api = api
        self.proxy = proxy
        return
    
    # 请求发送
    def request(
        self,
        method: Literal['get', 'post'],
        *,
        params: Optional[dict[str, Any]]=None,
        data: Optional[dict[str, Any]]=None,
        files: Optional[dict[str, tuple[str, BufferedReader, Optional[str]]]]=None
    ) -> Response:
        '''请求发送

        参数:
            method (Literal[&#39;get&#39;, &#39;post&#39;]): API 调用方法
            
            params (Optional[dict[str, Any]], optional): 调用参数
            
            data (Optional[dict[str, Any]], optional): 调用参数，将以 `json` 字符串形式发送
            
            files (Optional[dict[str, tuple[str, BufferedReader, Optional[str]]]], optional): 发送文件参数

        返回:
            Response: 收到的响应
        '''
        # 处理接收到的 API
        if self.api.startswith('http://') or self.api.startswith('https://'):
            self.api = self.api
        else:
            self.api = 'https://api.ayachan.fun/' + self.api
        # 构建一个请求体
        request = Request(
            method,
            self.api,
            params=params,
            data=(
                cast(dict, dumps(data)) if data is not None and 'sonolus' not in self.api
                else data
            ),
            files=files,
            headers={
                'Content-Type': 'multipart/form-data; boundary=----WebKitFormBoundaryxjpDgorEXUAUIbBN'
            } if 'sonolus' in self.api else None
        )
        # 构建代理服务器字典
        if self.proxy is not None:
            proxies = {'http://': self.proxy, 'https://': self.proxy}
        else:
            proxies = None
        
        # 发送请求并获取响应
        with Client(proxies=cast(dict, proxies)) as client:
            response = client.send(request)
            client.close()
        
        # 处理接收到的响应
        response.raise_for_status()
        return response



'''`bestdori.events`

BanG Dream! 活动相关操作'''
from typing import Optional, Literal, Any

from .post import get_list
from .utils.utils import API, ASSETS
from .utils.network import Api, Assets
from .eventarchives import EventArchive
from .exceptions import (
    ServerNotAvailableError,
    EventHasNoStampError,
    AssetsNotExistError,
    EventNotExistError
)

# 获取总活动信息
def get_all(index: Literal[0, 5, 6]=5, proxy: Optional[str]=None) -> dict[str, dict[str, Any]]:
    '''获取总活动信息

    参数:
        index (Literal[0, 5, 6], optional): 指定获取哪种 `all.json`
            `0`: 仅获取所有已有活动 ID `all.0.json`
            `5`: 获取所有已有活动的简洁信息 `all.5.json`
            `6`: 获取所有已有活动的简洁信息 `all.6.json`
        
        proxy (Optional[str], optional): 代理服务器

    返回:
        dict[str, dict[str, Any]]: 获取到的总活动信息
    '''
    return Api(API['events']['all'].format(index), proxy=proxy).request('get').json()

# 活动类
class Event:
    '''活动类

    参数:
        id_ (int): 活动 ID
        
        proxy (Optional[str], optional): 代理服务器
    '''
    # 初始化
    def __init__(self, id_: int, proxy: Optional[str]=None) -> None:
        '''活动类

        参数:
            id_ (int): 活动 ID
            
            proxy (Optional[str], optional): 代理服务器
        '''
        self.id: int = id_
        '''活动 ID'''
        self.archive: EventArchive = EventArchive(self.id, self.proxy)
        '''活动数据'''
        self._info: dict[str, Any] = {}
        '''活动信息'''
        self.proxy: Optional[str] = proxy
        '''代理服务器'''
        # 检测 ID 是否存在
        all_id = get_all(0, proxy=proxy)
        if not str(id_) in all_id.keys():
            raise EventNotExistError(id_)
        return
    
    # 获取活动信息
    def get_info(self) -> dict[str, Any]:
        '''获取活动信息

        返回:
            dict[str, Any]: 活动详细信息
        '''
        if len(self._info) <= 0:
            # 如果没有活动信息存储
            response = Api(
                API['events']['info'].format(self.id), proxy=self.proxy
            ).request('get')
            self._info = dict(response.json())
        return self._info
    
    # 获取活动评论
    def get_comment(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_ASC'
    ) -> dict[str, Any]:
        '''获取活动评论

        参数:
            limit (int, optional): 展示出的评论数，默认为 20
            
            offset (int, optional): 忽略前面的 `offset` 条评论，默认为 0
            
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 排序顺序，默认时间顺序

        返回:
            dict[str, Any]: 搜索结果
            ```python
            result: bool # 是否有响应
            count: int # 搜索到的评论总数
            posts: list[dict[str, Any]] # 列举出的评论
            ```
        '''
        return get_list(
            proxy=self.proxy,
            category_name='EVENT_COMMENT',
            category_id=str(self.id),
            order=order,
            limit=limit,
            offset=offset
        )
    
    # 获取活动标题
    @property
    def name(self) -> str:
        '''获取活动标题

        返回:
            str: 活动标题
        '''
        info = self.get_info()
        # 获取 eventName 数据
        if (event_name := info.get('eventName', None)) is None:
            raise Exception('无法获取活动标题。')
        # 获取第一个非 None 活动标题
        try:
            return next(filter(lambda x: x is not None, event_name))
        except StopIteration:
            raise Exception('无法获取活动标题。')
    
    # 获取活动默认服务器
    @property
    def server(self) -> Literal['jp', 'en', 'tw', 'cn', 'kr']:
        '''获取活动默认服务器

        返回:
            Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]: 歌曲所在服务器
        '''
        info = self.get_info()
        # 获取 startAt 数据
        if (start_at := info.get('startAt', None)) is None:
            raise Exception('无法获取活动起始时间。')
        # 根据 startAt 数据判断服务器
        if start_at[0] is not None: return 'jp'
        elif start_at[1] is not None: return 'en'
        elif start_at[2] is not None: return 'tw'
        elif start_at[3] is not None: return 'cn'
        elif start_at[4] is not None: return 'kr'
        else:
            raise Exception('无法获取活动所在服务器。')
    
    # 获取活动缩略图图像
    def get_banner(self, server: Literal['jp', 'en', 'tw', 'cn', 'kr']) -> bytes:
        '''获取活动缩略图图像

        参数:
            server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器

        返回:
            bytes: 活动缩略图图像字节数据 `bytes`
        '''
        # 获取活动数据包名称
        info = self.get_info()
        if (asset_bundle_name := info.get('assetBundleName', None)) is None:
            raise ValueError('无法获取活动数据包名称。')
        # 判断服务器
        if (start_at := info.get('startAt', None)) is None:
            raise ValueError('无法获取活动起始时间。')
        SERVERS = ['jp', 'en', 'tw', 'cn', 'kr']
        index = SERVERS.index(server)
        if start_at[index] is None:
            raise ServerNotAvailableError(f'活动 {self.name}', server)
        return Assets(
            ASSETS['event']['banner'].format(
                asset_bundle_name=asset_bundle_name
            ), server, self.proxy
        ).get()
    
    # 获取活动 logo 图像
    def get_logo(self, server: Literal['jp', 'en', 'tw', 'cn', 'kr']) -> bytes:
        '''获取活动 logo 图像

        参数:
            server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器

        返回:
            bytes: 活动 logo 图像字节数据 `bytes`
        '''
        # 获取活动数据包名称
        info = self.get_info()
        if (asset_bundle_name := info.get('assetBundleName', None)) is None:
            raise ValueError('无法获取活动数据包名称。')
        # 判断服务器
        if (start_at := info.get('startAt', None)) is None:
            raise ValueError('无法获取活动起始时间。')
        SERVERS = ['jp', 'en', 'tw', 'cn', 'kr']
        index = SERVERS.index(server)
        if start_at[index] is None:
            raise ServerNotAvailableError(f'活动 {self.name}', server)
        return Assets(
            ASSETS['event']['logo'].format(
                asset_bundle_name=asset_bundle_name
            ), server, self.proxy
        ).get()

    # 获取活动主界面图像
    def get_topscreen(
        self,
        server: Literal['jp', 'en', 'tw', 'cn', 'kr'],
        type_: Literal['bg', 'trim']
    ) -> bytes:
        '''获取活动主界面图像

        参数:
            server (Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]): 指定服务器
            
            type_ (Literal[&#39;bg&#39;, &#39;trim&#39;]): 图像类型
                `bg`: 背景图像
                `trim`: 角色图像

        返回:
            bytes: 活动主界面图像字节数据 `bytes`
        '''
        # 获取活动数据包名称
        info = self.get_info()
        if (asset_bundle_name := info.get('assetBundleName', None)) is None:
            raise ValueError('无法获取活动数据包名称。')
        # 判断服务器
        if (start_at := info.get('startAt', None)) is None:
            raise ValueError('无法获取活动起始时间。')
        SERVERS = ['jp', 'en', 'tw', 'cn', 'kr']
        index = SERVERS.index(server)
        if start_at[index] is None:
            raise ServerNotAvailableError(f'活动 {self.name}', server)
        return Assets(
            ASSETS['event']['topscreen'].format(
                asset_bundle_name=asset_bundle_name, type=type_
            ), server, self.proxy
        ).get()
    
    # 获取活动奖励稀有表情
    def get_stamp(self) -> bytes:
        '''获取活动奖励稀有表情

        返回:
            bytes: 活动奖励稀有表情字节数据 `bytes`
        '''
        info = self.get_info()
        # 获取活动点数奖励列表
        if (point_rewards := info.get('pointRewards', None)) is None:
            raise Exception('获取活动奖励失败。')
        # 获取第一个非 None 奖励列表
        try:
            point_reward = next(filter(lambda x: x is not None, point_rewards))
        except StopIteration:
            raise Exception('获取活动奖励失败。')
        # 获取 rewardType 为 stamp 的活动奖励
        try:
            reward = next(filter(lambda x: x['rewardType'] == 'stamp', point_reward))
        except StopIteration:
            raise EventHasNoStampError(self.id)
        stamp_id = reward['rewardId']
        # 获取全部贴纸资源
        stamps = Api(
            API['all']['stamps'].format(index='2'), proxy=self.proxy
        ).request('get').json()
        if (stamp := stamps.get(stamp_id, None)) is None:
            raise AssetsNotExistError(f'贴纸 {stamp_id}')
        # 获取贴纸资源
        image_name = stamp['imageName']
        return Assets(
            ASSETS['stamp']['get'].format(image_name=image_name), self.server, self.proxy
        ).get()

    # 获取排名分数线
    def get_top(
        self,
        server: Optional[Literal[0, 1, 2, 3, 4]]=None,
        mid: Literal['0']='0',
        latest: Literal['1']='1'
    ) -> dict[str, list[dict[str, Any]]]:
        '''获取排名分数线

        参数:
            server (Optional[Literal[0, 1, 2, 3, 4]], optional): 指定服务器
                `0`: 日服
                `1`: 英服
                `2`: 台服
                `3`: 国服
                `4`: 韩服
            
            mid (Literal[&#39;0&#39;], optional): 指定是否为中间分数线，默认为 `0`
            
            latest (Literal[&#39;1&#39;], optional): 指定是否为最终分数线，默认为 `1`

        返回:
            dict[str, list[dict[str, Any]]]: 排名分数线数据
        '''
        if server is None:
            if self.server == 'jp': server = 0
            elif self.server == 'en': server = 1
            elif self.server == 'tw': server = 2
            elif self.server == 'cn': server = 3
            elif self.server == 'kr': server = 4
            else: raise Exception('无法获取活动服务器。')
        return self.archive.get_top(server, mid, latest)



'''`bestdori.user`

BanG Dream! 歌曲相关操作'''
from httpx._models import Cookies
from typing import Optional, Literal, Any

from .post import get_list
from .utils.utils import API
from .utils.network import Api

# 用户类
class User:
    '''用户类

    参数:
        username (str): 用户名
        
        proxy (Optional[str], optional): 代理服务器
    '''
    # 初始化
    def __init__(self, username: str, proxy: Optional[str]=None) -> None:
        '''用户类

        参数:
            username (str): 用户名
            
            proxy (Optional[str], optional): 代理服务器
        '''        
        self.username: str = username
        '''用户名'''
        self._info: dict[str, Any] = {}
        '''用户信息'''
        self.proxy: Optional[str] = proxy
        '''代理服务器'''
        return
    
    # 获取用户详细信息
    def get_info(self) -> dict[str, Any]:
        '''获取用户详细信息

        返回:
            dict[str, Any]: 用户详细信息
        '''
        if len(self._info) <= 0:
            # 若无用户信息存储
            response = Api(
                API['user']['info'], self.proxy
            ).request('get', params={'username': self.username})
            self._info = dict(response.json())
        return self._info
    
    # 获取用户帖子
    def get_posts(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_DESC'
    ) -> dict[str, Any]:
        '''获取用户帖子

        参数:
            limit (int, optional): 展示出的帖子数，默认 20
            
            offset (int, optional): 忽略前面的 `offset` 个帖子，默认 0
            
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 帖子排序，默认时间倒序

        返回:
            dict[str, Any]: 获取结果
            ```python
            result: bool # 是否有响应
            count: int # 搜索到的帖子总数
            posts: list[dict[str, Any]] # 列举出的帖子
            ```
        '''
        return get_list(
            self.proxy, username=self.username, limit=limit, offset=offset, order=order
        )
    
    # 获取用户谱面
    def get_charts(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_DESC'
    ) -> dict[str, Any]:
        '''获取用户谱面

        参数:
            limit (int, optional): 展示出的谱面数，默认 20
            
            offset (int, optional): 忽略前面的 `offset` 个谱面，默认 0
            
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 谱面排序，默认时间倒序

        返回:
            dict[str, Any]: 获取结果
            ```python
            result: bool # 是否有响应
            count: int # 搜索到的谱面总数
            posts: list[dict[str, Any]] # 列举出的谱面
            ```
        '''
        return get_list(
            self.proxy,
            username=self.username,
            category_name='SELF_POST',
            category_id='chart',
            limit=limit,
            offset=offset,
            order=order
        )
    
    # 获取用户文本帖子
    def get_texts(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_DESC'
    ) -> dict[str, Any]:
        '''获取用户文本帖子

        参数:
            limit (int, optional): 展示出的帖子数，默认 20
            
            offset (int, optional): 忽略前面的 `offset` 个帖子，默认 0
            
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 帖子排序，默认时间倒序

        返回:
            dict[str, Any]: 获取结果
            ```python
            result: bool # 是否有响应
            count: int # 搜索到的帖子总数
            posts: list[dict[str, Any]] # 列举出的帖子
            ```
        '''
        return get_list(
            self.proxy,
            username=self.username,
            category_name='SELF_POST',
            category_id='text',
            limit=limit,
            offset=offset,
            order=order
        )
    
    # 获取用户故事
    def get_storys(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_DESC'
    ) -> dict[str, Any]:
        '''获取用户故事

        参数:
            limit (int, optional): 展示出的故事数，默认 20
            
            offset (int, optional): 忽略前面的 `offset` 个故事，默认 0
            
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 故事排序，默认时间倒序

        返回:
            dict[str, Any]: 获取结果
            ```python
            result: bool # 是否有响应
            count: int # 搜索到的故事总数
            posts: list[dict[str, Any]] # 列举出的故事
            ```
        '''
        return get_list(
            self.proxy,
            username=self.username,
            category_name='SELF_POST',
            category_id='story',
            limit=limit,
            offset=offset,
            order=order
        )

# 自身用户类
class Me(User):
    '''自身用户类

    参数:
        username (str): 用户名
        
        password (str): 密码
        
        proxy (Optional[str], optional): 代理服务器
    '''
    # 初始化
    def __init__(self, username: str, password: str, proxy: Optional[str]=None) -> None:
        '''自身用户类

        参数:
            username (str): 用户名
            
            password (str): 密码
            
            proxy (Optional[str], optional): 代理服务器
        '''
        super().__init__(username, proxy)
        self.password: str = password
        '''密码'''
        self.cookies: Optional[Cookies] = None
        '''用户 Cookies'''
        self._me: dict[str, Any] = {}
        '''用户自我信息'''
        return
    
    # 用户登录
    @classmethod
    def login(cls, username: str, password: str, proxy: Optional[str]=None) -> 'Me':
        '''用户登录

        参数:
            username (str): 用户名
            
            password (str): 密码
            
            proxy (Optional[str], optional): 代理服务器

        返回:
            Me: 自身用户对象
        '''
        me = cls(username, password, proxy)
        response = Api(
            API['user']['login'], me.proxy
        ).request('post', data={'username': me.username, 'password': me.password})
        me.cookies = response.cookies
        me.me()
        return me
    
    # 获取用户自我信息
    def me(self) -> dict[str, Any]:
        '''获取用户自我信息

        返回:
            dict[str, Any]: 自我信息
        '''
        if len(self._me) == 0:
            response = Api(API['user']['me'], self.proxy).request('get', cookies=self.cookies)
            self._me = dict(response.json())
        return self._me



'''`bestdori.songs`

BanG Dream! 歌曲相关操作'''
from typing import Optional, Literal, Any
from requests.exceptions import HTTPError

from .charts import Chart
from .post import get_list
from .utils.utils import ASSETS, API
from .utils.network import Assets, Api
from .exceptions import (
    DiffNotExistError,
    SongNotExistError
)

# 获取总歌曲信息
def get_all(index: Literal[0, 5]=5, proxy: Optional[str]=None) -> dict[str, dict[str, Any]]:
    '''获取总歌曲信息

    参数:
        index (Literal[0, 5], optional): 指定获取哪种 `all.json`
            `0`: 仅获取所有已有歌曲 ID `all.0.json`
            `5`: 获取所有已有歌曲的简洁信息 `all.5.json`，默认为该项
        
        proxy (Optional[str], optional): 代理服务器

    返回:
        dict[str, dict[str, Any]]: 获取到的总歌曲信息
    '''
    return Api(API['songs']['all'].format(index=index), proxy=proxy).request('get').json()

# 歌曲类
class Song:
    '''歌曲类

    参数:
        id_ (int): 歌曲 ID
        
        proxy (Optional[str], optional): 代理服务器
    '''
    # 初始化
    def __init__(self, id_: int, proxy: Optional[str]=None) -> None:
        '''歌曲类

        参数:
            id_ (int): 歌曲 ID
            
            proxy (Optional[str], optional): 代理服务器
        '''
        self.id: int = id_
        '''歌曲 ID'''
        self._info: dict[str, Any] = {}
        '''歌曲信息'''
        self.proxy: Optional[str] = proxy
        '''代理服务器'''
        # 检测 ID 是否存在
        all_id = get_all(0, proxy=proxy)
        if not str(id_) in all_id.keys():
            raise SongNotExistError(id_)
        return
    
    # 获取歌曲信息
    def get_info(self) -> dict[str, Any]:
        '''获取歌曲信息

        返回:
            dict[str, Any]: 歌曲详细信息
        '''
        if len(self._info) <= 0:
            # 如果没有歌曲信息存储
            response = Api(
                API['songs']['info'].format(id=self.id), proxy=self.proxy
            ).request('get')
            self._info = dict(response.json())
        return self._info
    
    # 获取歌曲所在服务器
    @property
    def server(self) -> Literal['jp', 'en', 'tw', 'cn', 'kr']:
        '''获取歌曲所在服务器

        返回:
            Literal[&#39;jp&#39;, &#39;en&#39;, &#39;tw&#39;, &#39;cn&#39;, &#39;kr&#39;]: 歌曲所在服务器
        '''
        info = self.get_info()
        # 获取 publishedAt 数据
        if (published_at := info.get('publishedAt', None)) is None:
            raise Exception('无法获取歌曲发布时间。')
        # 根据 publishedAt 数据判断服务器
        if published_at[0] is not None: return 'jp'
        elif published_at[1] is not None: return 'en'
        elif published_at[2] is not None: return 'tw'
        elif published_at[3] is not None: return 'cn'
        elif published_at[4] is not None: return 'kr'
        else:
            raise Exception('无法获取歌曲服务器。')
    
    # 获取歌曲名称
    @property
    def name(self) -> str:
        '''获取歌曲名称

        返回:
            str: 歌曲名称
        '''
        info = self.get_info()
        # 获取 musicTitle 数据
        if (music_title := info.get('musicTitle', None)) is None:
            raise Exception('无法获取歌曲名称。')
        # 获取第一个非 None 歌曲名称
        try:
            return next(filter(lambda x: x is not None, music_title))
        except StopIteration:
            raise Exception('无法获取歌曲名称。')
    
    # 获取歌曲谱面
    def get_chart(
        self,
        diff: Literal['easy', 'normal', 'hard', 'expert', 'special']='expert'
    ) -> Chart:
        '''获取歌曲谱面

        参数:
            diff (Literal[&#39;easy&#39;, &#39;normal&#39;, &#39;hard&#39;, &#39;expert&#39;, &#39;special&#39;], optional): 难度名称

        返回:
            Chart: 获取到的谱面对象
        '''
        try:
            chart = Chart.get_chart(self.id, diff, self.proxy)
            return chart
        except HTTPError:
            # 难度不存在
            raise DiffNotExistError(diff)
    
    # 获取歌曲封面
    def get_jacket(self) -> list[bytes]:
        '''获取歌曲封面

        返回:
            bytes: 歌曲封面字节数据 `bytes`
        '''
        # 获取数据包序列号
        quotient, remainder = divmod(int(self.id), 10)
        if remainder == 0:
            index = self.id
        else:
            index = str((quotient + 1) * 10)
        
        info = self.get_info()
        if (jacket_image := info.get('jacketImage', None)) is None:
            raise Exception('歌曲封面资源未找到。')
        jacket: list[bytes] = []
        
        for image in jacket_image:
            jacket.append(
                Assets(
                    ASSETS['songs']['musicjacket'].format(
                        index=index, jacket_image=image
                    ), self.server, self.proxy
                ).get()
            )
        
        return jacket
    
    # 获取歌曲音频
    def get_bgm(self) -> bytes:
        '''获取歌曲音频

        返回:
            bytes: 歌曲音频字节数据 `bytes`
        '''
        return Assets(ASSETS['songs']['sound'].format(id=self.id), self.server, self.proxy).get()
    
    # 获取歌曲评论
    def get_comment(
        self,
        limit: int=20,
        offset: int=0,
        order: Literal['TIME_DESC', 'TIME_ASC']='TIME_ASC'
    ) -> dict[str, Any]:
        '''获取歌曲评论

        参数:
            limit (int, optional): 展示出的评论数，默认为 20
            
            offset (int, optional): 忽略前面的 `offset` 条评论，默认为 0
            
            order (Literal[&#39;TIME_DESC&#39;, &#39;TIME_ASC&#39;], optional): 排序顺序，默认时间顺序

        返回:
            dict[str, Any]: 搜索结果
            ```python
            result: bool # 是否有响应
            count: int # 搜索到的评论总数
            posts: list[dict[str, Any]] # 列举出的评论
            ```
        '''
        return get_list(
            proxy=self.proxy,
            category_name='SONG_COMMENT',
            category_id=str(self.id),
            order=order,
            limit=limit,
            offset=offset
        )



'''`bestdori.songmeta`

BanG Dream! 歌曲 Meta 相关操作'''
from typing import Optional, Literal, Any

from .utils.utils import API
from .utils.network import Api

# 获取总歌曲 Meta 信息
def get_all(index: Literal[5]=5, proxy: Optional[str]=None) -> dict[str, dict[str, Any]]:
    '''获取总歌曲 Meta 信息

    参数:
        index (Literal[5], optional): 指定获取哪种 `all.json`
            `5`: 获取所有已有歌曲 Meta 信息 `all.5.json`
        
        proxy (Optional[str], optional): 代理服务器

    返回:
        dict[str, dict[str, Any]]: 获取到的总歌曲 Meta 信息
    '''
    return Api(API['all']['meta'].format(index=index), proxy=proxy).request('get').json()
